@@ -1,6 +1,6 @@
 /*********************************************************************
  *
- *      Copyright (C) 2001-2002 Nathan Fiedler
+ *      Copyright (C) 2001-2004 Nathan Fiedler
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -16,47 +16,37 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
- * PROJECT:     JSwat
- * MODULE:      JSwat
- * FILE:        PathManager.java
- *
- * AUTHOR:      Nathan Fiedler
- *
- * REVISION HISTORY:
- *      Name    Date            Description
- *      ----    ----            -----------
- *      nf      08/06/01        Initial version
- *      nf      10/21/01        Fixed bug #259
- *      nf      11/25/01        Fixed bug #321
- *      nf      01/03/02        Fixed bug #375
- *      nf      02/20/02        Fixed bug #52
- *      nf      02/21/02        Fixed bug #306
- *      MvD     03/10/02        Use Session classpath and sourcepath
- *      nf      03/18/02        Fixed bug #411
- *      nf      04/25/02        Fixed bug #488 (sorta)
- *      nf      04/26/02        Fixed bug #490
- *      nf      05/12/02        Fixed bug #531
- *      nf      05/22/02        Fixing mistake made in 276
- *      nf      10/27/02        Fixed bug #575, 632
- *
- * DESCRIPTION:
- *      Defines the class responsible for managing classpath and
- *      sourcepath.
- *
- * $Id: PathManager.java 640 2002-10-27 22:54:26Z nfiedler $
+ * $Id: PathManager.java 1173 2004-02-27 20:51:07Z nfiedler $
  *
  ********************************************************************/
 
 package com.bluemarsh.jswat;
 
-import com.bluemarsh.config.JConfigure;
+import com.bluemarsh.jswat.event.SessionEvent;
+import com.bluemarsh.jswat.lang.ClassName;
+import com.bluemarsh.jswat.ui.UIAdapter;
 import com.sun.jdi.AbsentInformationException;
 import com.sun.jdi.PathSearchingVirtualMachine;
 import com.sun.jdi.ReferenceType;
 import com.sun.jdi.VirtualMachine;
-import java.io.*;
-import java.util.*;
-import java.util.zip.*;
+import com.sun.jdi.Location;
+import java.io.File;
+import java.io.IOException;
+import java.io.InputStream;
+import java.util.Enumeration;
+import java.util.Hashtable;
+import java.util.Iterator;
+import java.util.List;
+import java.util.StringTokenizer;
+import java.util.zip.ZipEntry;
+import java.util.zip.ZipFile;
+import org.gjt.jclasslib.io.ClassFileReader;
+import org.gjt.jclasslib.structures.AttributeInfo;
+import org.gjt.jclasslib.structures.ClassFile;
+import org.gjt.jclasslib.structures.CPInfo;
+import org.gjt.jclasslib.structures.InvalidByteCodeException;
+import org.gjt.jclasslib.structures.attributes.SourceFileAttribute;
+import org.gjt.jclasslib.structures.constants.ConstantUtf8Info;
 
 /**
  * Class PathManager is responsible for managing the classpath and
@@ -69,38 +59,101 @@ import java.util.zip.*;
  * @author  Nathan Fiedler
  * @author  Marko van Dooren
  */
-public class PathManager extends DefaultManager {
+public class PathManager implements Manager {
     /** Session that owns us. */
-    protected Session owningSession;
-    /** Table of SourceSource objects, keyed by the classname. */
-    protected Hashtable classnameSources;
+    private Session owningSession;
+    /** Generator of source objects. */
+    private SourceFactory sourceFactory;
 
     /**
      * Constructs a PathManager object.
      */
     public PathManager() {
-        classnameSources = new Hashtable();
+        sourceFactory = SourceFactory.getInstance();
     } // PathManager
 
     /**
-     * Turn the package name into a file path using simple character
-     * subsitution.
+     * Called when the Session has activated. This occurs when the
+     * debuggee has launched or has been attached to the debugger.
      *
-     * @param  classname  fully-qualified name of the class, possibly
-     *                    including an inner-class specification.
-     * @return  path and filename of source file.
+     * @param  sevt  session event.
      */
-    protected String classnameToFilename(String classname) {
-        int dollar = classname.indexOf('$');
-        if (dollar > 0) {
-            // Drop the inner-class specifier.
-            classname = classname.substring(0, dollar);
+    public void activated(SessionEvent sevt) {
+    } // activated
+
+    /**
+     * Called when the Session is about to be closed.
+     *
+     * @param  sevt  session event.
+     */
+    public void closing(SessionEvent sevt) {
+    } // closing
+
+    /**
+     * Called when the Session has deactivated. The debuggee VM is no
+     * longer connected to the Session.
+     *
+     * @param  sevt  session event.
+     */
+    public void deactivated(SessionEvent sevt) {
+    } // deactivated
+
+    /**
+     * Attempt to determine the name of the package that the given path
+     * belongs to, using the sourcepath and classpath settings.
+     *
+     * @param  path  full path to source file.
+     * @return  the package this source file belongs to; empty string
+     *          if there is no actual package; null if error or path
+     *          is not in the sourcepath or classpath.
+     */
+    public String findPackageName(String path) {
+        if (path == null || path.length() == 0) {
+            return null;
         }
-        classname = classname.replace('.', File.separatorChar);
-        JConfigure config = JSwat.instanceOf().getJConfigure();
-        classname += config.getProperty("files.defaultExtension");
-        return classname;
-    } // classnameToFilename
+
+        // Must canonicalize the path first.
+        try {
+            path = new File(path).getCanonicalPath();
+        } catch (IOException ioe) {
+            return null;
+        }
+        // Search the sourcepath.
+        String[] paths = getSourcePath();
+        for (int ii = 0; ii < paths.length; ii++) {
+            String apath = paths[ii];
+            if (path.startsWith(apath)) {
+                // Remove the prefix and convert to a package name.
+                if (apath.length() < path.length()) {
+                    path = path.substring(apath.length() + 1);
+                    return path.replace(File.separatorChar, '.');
+                } else if (apath.length() == path.length()) {
+                    return "";
+                } else {
+                    return null;
+                }
+            }
+        }
+
+        // Search the classpath.
+        paths = getClassPath();
+        for (int ii = 0; ii < paths.length; ii++) {
+            String apath = paths[ii];
+            if (path.startsWith(apath)) {
+                // Remove the prefix and convert to a package name.
+                if (apath.length() < path.length()) {
+                    path = path.substring(apath.length() + 1);
+                    return path.replace(File.separatorChar, '.');
+                } else if (apath.length() == path.length()) {
+                    return "";
+                } else {
+                    return null;
+                }
+            }
+        }
+
+        return null;
+    } // findPackageName
 
     /**
      * Returns the array of classpath directories, if any.
@@ -150,10 +203,7 @@ public class PathManager extends DefaultManager {
         }
 
         String classpath = owningSession.getProperty("classpath");
-        if (classpath == null) {
-            classpath = "";
-        }
-        return classpath;
+        return classpath == null ? "" : classpath;
     } // getClassPathAsString
 
     /**
@@ -180,21 +230,257 @@ public class PathManager extends DefaultManager {
      */
     public String getSourcePathAsString() {
         String sourcepath = owningSession.getProperty("sourcepath");
-        if (sourcepath == null) {
-            sourcepath = "";
-        }
-        return sourcepath;
+        return sourcepath == null ? "" : sourcepath;
     } // getSourcePathAsString
 
     /**
-     * Called after the Session has instantiated this mananger.
-     * To avoid problems with circular dependencies between managers,
-     * iniitialize data members before calling
-     * <code>Session.getManager()</code>.
+     * Searches for the .class file of the given class and returns
+     * a SourceSource representing that .class file.
      *
-     * @param  session  Session initializing this manager.
+     * @param  clazz  class for which to find class file.
+     * @return  object representing the desired class file, or null
+     *          if class file not found.
      */
-    public void init(Session session) {
+    public SourceSource mapClass(ReferenceType clazz) {
+        return mapClass(clazz.name());
+    } // mapClass
+
+    /**
+     * Searches for the .class file of the given name and returns
+     * a SourceSource representing that .class file.
+     *
+     * @param  name  name of class for which to find class file.
+     * @return  object representing the desired class file, or null
+     *          if class file not found.
+     */
+    public SourceSource mapClass(String name) {
+        String filename = name.replace('.', File.separatorChar) + ".class";
+        SourceSource src = null;
+        String[] classpathArray = getClassPath();
+        if (classpathArray != null) {
+            for (int ii = 0; ii < classpathArray.length; ii++) {
+                src = mapFile(classpathArray[ii], filename);
+                if (src != null) {
+                    break;
+                }
+            }
+        }
+        return src;
+    } // mapClass
+
+    /**
+     * Returns a SourceSource for the relatively named file (file path is
+     * relative to sourcepath or classpath), scanning the sourcepath and
+     * classpath to find the absolute location of the file.
+     *
+     * @param  filename  relative path and name of file.
+     * @return  matching file source, if found.
+     */
+    public SourceSource mapFile(String filename) {
+        String[] sourcepathArray = getSourcePath();
+        // Scan through the sourcepath list, if available.
+        if (sourcepathArray != null) {
+            for (int ii = 0; ii < sourcepathArray.length; ii++) {
+                SourceSource src = mapFile(sourcepathArray[ii], filename);
+                if (src != null) {
+                    return src;
+                }
+            }
+        }
+
+        String[] classpathArray = getClassPath();
+        // Scan through the classpath list, if available.
+        if (classpathArray != null) {
+            for (int ii = 0; ii < classpathArray.length; ii++) {
+                SourceSource src = mapFile(classpathArray[ii], filename);
+                if (src != null) {
+                    return src;
+                }
+            }
+        }
+
+        // Did not find a matching source file.
+        return null;
+    } // mapFile
+
+    /**
+     * Look for the file in the given classpath or sourcepath entry. This
+     * method deals with zip and jar archives, as well as the usual
+     * directory paths.
+     *
+     * @param  path      source or classpath entry.
+     * @param  filename  path and name of file to look for.
+     * @return  source file, if found, null otherwise.
+     */
+    protected SourceSource mapFile(String path, String filename) {
+        if (path == null) {
+            return null;
+        }
+
+        File f = new File(path);
+        if (f.isFile() && (path.endsWith(".zip") || path.endsWith(".jar"))) {
+            // Zip/jar path entry.
+            ZipFile zipFile = null;
+            try {
+                zipFile = new ZipFile(path);
+            } catch (IOException ioe) {
+                return null;
+            }
+            Enumeration entries = zipFile.entries();
+            while (entries.hasMoreElements()) {
+                ZipEntry zipEntry = (ZipEntry) entries.nextElement();
+                String entryName = zipEntry.getName();
+                // Convert the name to the local file system form so we
+                // can compare it to the filename argument.
+                entryName = new File(entryName).getPath();
+                if (entryName.equals(filename)) {
+                    return sourceFactory.create(zipFile, zipEntry);
+                }
+            }
+
+        } else {
+            // Directory path entry.
+            f = new File(path, filename);
+            if (f.exists()) {
+                return sourceFactory.create(path, filename, this);
+            }
+        }
+
+        return null;
+    } // mapFile
+
+    /**
+     * Return a SourceSource corresponding to the fully-qualified class
+     * name. Return null if the source was not found.
+     *
+     * @param  classname  fully-qualified class name.
+     * @return  source containing the desired location.
+     * @throws  IOException
+     *          if an I/O error occurred.
+     */
+    public SourceSource mapSource(String classname) throws IOException {
+        // First try the primitive means of finding the source.
+        String filename = ClassName.toFilename(classname);
+        SourceSource source = mapFile(filename);
+        if (source == null) {
+            // Now try using jclasslib to get the SourceFile attribute from
+            // the .class file itself.
+            SourceSource clazz = mapClass(classname);
+            if (clazz != null) {
+                InputStream is = clazz.getInputStream();
+                String sourcename = null;
+                try {
+                    ClassFile classFile =
+                        ClassFileReader.readFromInputStream(is);
+                    AttributeInfo[] ainfo = classFile.getAttributes();
+                    for (int ii = 0; ii < ainfo.length; ii++) {
+                        if (ainfo[ii] instanceof SourceFileAttribute) {
+                            SourceFileAttribute sfa =
+                                (SourceFileAttribute) ainfo[ii];
+                            int index = sfa.getSourcefileIndex();
+                            ConstantUtf8Info cpinfo = (ConstantUtf8Info)
+                                classFile.getConstantPoolEntry(
+                                    index, ConstantUtf8Info.class);
+                            sourcename = new String(cpinfo.getString());
+                            break;
+                        }
+                    }
+
+                } catch (InvalidByteCodeException ibce) {
+                    owningSession.getUIAdapter().showMessage(
+                        UIAdapter.MESSAGE_WARNING,
+                        Bundle.getString("PathManager.badbytes") + ' '
+                        + classname);
+                } catch (IOException ioe) {
+                    owningSession.getUIAdapter().showMessage(
+                        UIAdapter.MESSAGE_WARNING,
+                        Bundle.getString("PathManager.ioExc") + ' ' + ioe);
+                }
+                if (sourcename != null) {
+                    filename = ClassName.toFilename(classname, sourcename);
+                    source = mapFile(filename);
+                }
+            }
+        }
+
+        if (source == null) {
+            // Maybe we can find the class and make a bytecode view of it.
+            VirtualMachine vm = owningSession.getVM();
+            if (vm != null) {
+                List classes = vm.classesByName(classname);
+                if (classes.size() == 1) {
+                    ReferenceType clazz = (ReferenceType) classes.get(0);
+                    source = sourceFactory.create(clazz);
+                }
+            }
+        }
+        return source;
+    } // mapSource
+
+    /**
+     * Return a SourceSource corresponding to the given location.
+     * Return null if the source was not found. This method is preferred
+     * over the other mapSource() methods because the Location provides
+     * the most accurate information.
+     *
+     * @param  location  location for which to find source file.
+     * @return  source containing the desired location.
+     * @throws  IOException
+     *          if an I/O error occurred.
+     */
+    public SourceSource mapSource(Location location) throws IOException {
+        try {
+            String source = location.sourceName();
+            String classname = location.declaringType().name();
+            String filename = ClassName.toFilename(classname, source);
+            SourceSource src = mapFile(filename);
+            if (src != null) {
+                return src;
+            }
+        } catch (AbsentInformationException aie) {
+            // fall through...
+        }
+
+        // Fall back to using only the class information.
+        return mapSource(location.declaringType());
+    } // mapSource
+
+    /**
+     * Return a SourceSource corresponding to the given class. Return
+     * null if the source was not found. This method is preferred over
+     * using the mapSource(String) method, but is not as good as the
+     * mapSource(Location) method.
+     *
+     * @param  clazz  class for which to find source file.
+     * @return  source containing the desired location.
+     * @throws  IOException
+     *          if an I/O error occurred.
+     */
+    public SourceSource mapSource(ReferenceType clazz) throws IOException {
+        String classname = clazz.name();
+        String filename;
+        // Try to use the source filename as given by the class.
+        try {
+            String srcname = clazz.sourceName();
+            filename = ClassName.toFilename(classname, srcname);
+        } catch (AbsentInformationException aie) {
+            filename = ClassName.toFilename(classname);
+        }
+
+        SourceSource src = mapFile(filename);
+        if (src == null) {
+            src = sourceFactory.create(clazz);
+        }
+        return src;
+    } // mapSource
+
+    /**
+     * Called after the Session has added this listener to the Session
+     * listener list.
+     *
+     * @param  session  the Session.
+     */
+    public void opened(Session session) {
         owningSession = session;
         // This is the end-all-be-all solution for handling the paths:
 
@@ -214,190 +500,20 @@ public class PathManager extends DefaultManager {
             classpath = System.getProperty("java.class.path");
             setClassPath(classpath);
         }
-    } // init
+    } // opened
 
     /**
-     * Searches for the .class file of the given class and returns
-     * a SourceSource representing that .class file.
+     * Called when the debuggee is about to be resumed.
      *
-     * @param  clazz    class for which to find class file.
-     * @return  object representing the desired class file, or null
-     *          if class file not found.
+     * @param  sevt  session event.
      */
-    public SourceSource mapClass(ReferenceType clazz) {
-        String filename = clazz.name().replace('.', File.separatorChar) +
-            ".class";
-
-        SourceSource src = null;
-
-        String[] classpathArray = getClassPath();
-        if (classpathArray != null) {
-            for (int i = 0; i < classpathArray.length; i++) {
-                src = mapSource0(classpathArray[i], filename);
-                if (src != null) {
-                    break;
-                }
-            }
-        }
-
-        return src;
-    } // mapClass
-
-    /**
-     * Return a SourceSource corresponding to the fully-qualified class
-     * name. Return null if the source was not found.
-     *
-     * @param  classname  fully-qualified class name.
-     * @return  source containing the desired location.
-     * @exception  IOException
-     *             Thrown if an I/O error occurred.
-     */
-    public SourceSource mapSource(String classname) throws IOException {
-        // Check in the classname/File look-up table to see if
-        // we've already mapped this classname to a file.
-        SourceSource source = (SourceSource) classnameSources.get(classname);
-        if (source != null) {
-            return source;
-        }
-
-        // Use the primitive means of finding the source.
-        String filename = classnameToFilename(classname);
-        return mapSourceLow(filename, classname);
-    } // mapSource
-
-    /**
-     * Return a SourceSource corresponding to the given class.
-     * Return null if the source was not found.
-     *
-     * @param  clazz    class for which to find source file.
-     * @return  source containing the desired location.
-     * @exception  IOException
-     *             Thrown if an I/O error occurred.
-     */
-    public SourceSource mapSource(ReferenceType clazz) throws IOException {
-        // Check in the classname/SourceSource look-up table to see if
-        // we have already mapped this classname to a source.
-        String classname = clazz.name();
-        SourceSource source = (SourceSource) classnameSources.get(classname);
-        if (source != null) {
-            return source;
-        }
-
-        String filename = classnameToFilename(classname);
-
-        // Try to use the source filename as given by the class.
-        try {
-            // Get the source name first so the exception gets thrown
-            // now rather than after we modify the filename variable.
-            String srcname = clazz.sourceName();
-            int bsi = srcname.lastIndexOf('\\');
-            int fsi = srcname.lastIndexOf('/');
-            // Work-around for bug 4404985 where SourceFile has path.
-            if (bsi > -1) {
-                srcname = srcname.substring(bsi + 1);
-            } else if (fsi > -1) {
-                srcname = srcname.substring(fsi + 1);
-            }
-            int lastbit = filename.lastIndexOf(File.separatorChar);
-            if (lastbit > -1) {
-                filename = filename.substring(0, lastbit);
-                filename = filename + File.separator + srcname;
-            } else {
-                // Class without a path, just use the source name.
-                filename = srcname;
-            }
-        } catch (AbsentInformationException aie) {
-            // If this happens, this method ends up being the same
-            // as the version that takes a String argument.
-        }
-
-        return mapSourceLow(filename, classname);
-    } // mapSource
-
-    /**
-     * Looks for a matching entry in either the classpath or sourcepath.
-     *
-     * @param  filename   name of file to look for.
-     * @param  classname  name of class for caching result.
-     * @return  matching source, if found.
-     */
-    protected SourceSource mapSourceLow(String filename, String classname) {
-        String[] sourcepathArray = getSourcePath();
-        // Scan through the sourcepath list, if available.
-        if (sourcepathArray != null) {
-            for (int i = 0; i < sourcepathArray.length; i++) {
-                SourceSource src = mapSource0(sourcepathArray[i], filename);
-                if (src != null) {
-                    // We found it, cache it and return the source.
-                    classnameSources.put(classname, src);
-                    return src;
-                }
-            }
-        }
-
-        String[] classpathArray = getClassPath();
-        // Scan through the classpath list, if available.
-        if (classpathArray != null) {
-            for (int i = 0; i < classpathArray.length; i++) {
-                SourceSource src = mapSource0(classpathArray[i], filename);
-                if (src != null) {
-                    // We found it, cache it and return the source.
-                    classnameSources.put(classname, src);
-                    return src;
-                }
-            }
-        }
-
-        // Did not find a matching source file.
-        return null;
-    } // mapSourceLow
-
-    /**
-     * Look for the file in the given class or sourcepath entry.
-     * This method deals with zip and jar archives, as well as the
-     * usual directory paths.
-     *
-     * @param  path      source or classpath entry.
-     * @param  filename  name of file to look for.
-     */
-    protected SourceSource mapSource0(String path,
-                                      String filename) {
-
-        if (path.endsWith(".zip") || path.endsWith(".jar")) {
-            // Zip/jar path entry.
-            ZipFile zipFile = null;
-            try {
-                zipFile = new ZipFile(path);
-            } catch (IOException ioe) {
-                return null;
-            }
-            Enumeration enum = zipFile.entries();
-            while (enum.hasMoreElements()) {
-                ZipEntry zipEntry = (ZipEntry) enum.nextElement();
-                String entryName = zipEntry.getName();
-                // Convert the name to the local file system form so we
-                // can compare it to the filename argument.
-                entryName = new File(entryName).getPath();
-                if (entryName.equals(filename)) {
-                    return new ZipSource(zipFile, zipEntry);
-                }
-            }
-
-        } else {
-            // Directory path entry.
-            File file = new File(path, filename);
-            if (file.exists()) {
-                return new FileSource(file);
-            }
-        }
-
-        return null;
-    } // mapSource0
+    public void resuming(SessionEvent sevt) {
+    } // resuming
 
     /**
      * Sets the classpath this source manager uses. The previous
-     * classpath is discarded in favor of the new one. The classpath
-     * is used together with the sourcepath, if any.
+     * classpath is discarded in favor of the new one. The classpath is
+     * used together with the sourcepath, if any.
      *
      * @param  classpath  Classpath for VM.
      */
@@ -423,42 +539,10 @@ public class PathManager extends DefaultManager {
         owningSession.setProperty("classpath", sb.toString());
     } // setClassPath
 
-    /**
-     * Sets the classpath this source manager uses. The previous
-     * classpath is discarded in favor of the new one. The class
-     * path is used together with the sourcepath, if any.
-     *
-     * @param  list  List of String classpath entries.
-     */
-    public void setClassPath(List list) {
-        int size = list.size();
-        if (size == 0) {
-            // Special case of empty list.
-            owningSession.setProperty("classpath", "");
-        } else {
-            // Turn the list into a String so we can save it.
-            StringBuffer sb = new StringBuffer(80);
-            Iterator iter = list.iterator();
-            while (iter.hasNext()) {
-                String entry = (String) iter.next();
-                File f = new File(entry);
-                try {
-                    sb.append(f.getCanonicalPath());
-                } catch (IOException ioe) {
-                    sb.append(entry);
-                }
-                sb.append(File.pathSeparator);
-            }
-            // Remove the last path separator.
-            sb.deleteCharAt(sb.length() - 1);
-            owningSession.setProperty("classpath", sb.toString());
-        }
-    } // setClassPath
-
     /**
      * Sets the sourcepath this source manager uses. The previous
-     * sourcepath is discarded in favor of the new one. The source
-     * path is used together with the classpath, if any.
+     * sourcepath is discarded in favor of the new one. The source path
+     * is used together with the classpath, if any.
      *
      * @param  sourcepath  Sourcepath for VM.
      */
@@ -483,4 +567,12 @@ public class PathManager extends DefaultManager {
         }
         owningSession.setProperty("sourcepath", sb.toString());
     } // setSourcePath
+
+    /**
+     * Called when the debuggee has been suspended.
+     *
+     * @param  sevt  session event.
+     */
+    public void suspended(SessionEvent sevt) {
+    } // suspended
 } // PathManager
