@@ -1,6 +1,6 @@
 /*********************************************************************
  *
- *      Copyright (C) 2001 Nathan Fiedler
+ *      Copyright (C) 2001-2003 Nathan Fiedler
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -28,21 +28,21 @@
  *      nf      03/24/01        Initial version
  *      nf      06/11/01        Added group and breakpoint iterators
  *
- * DESCRIPTION:
- *      Defines the breakpoint group class.
- *
- * $Id: BreakpointGroup.java 629 2002-10-26 23:03:26Z nfiedler $
+ * $Id: BreakpointGroup.java 1146 2004-02-20 03:41:14Z nfiedler $
  *
  ********************************************************************/
 
 package com.bluemarsh.jswat.breakpoint;
 
 import com.bluemarsh.jswat.Session;
-import java.io.Serializable;
+import com.bluemarsh.jswat.ui.UIAdapter;
+import java.text.MessageFormat;
 import java.util.Iterator;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.NoSuchElementException;
+import java.util.prefs.BackingStoreException;
+import java.util.prefs.Preferences;
 
 /**
  * Class BreakpointGroup acts as a container for breakpoints.
@@ -50,7 +50,7 @@ import java.util.NoSuchElementException;
  *
  * @author  Nathan Fiedler
  */
-public class BreakpointGroup implements Serializable {
+public class BreakpointGroup {
     /** Name of our breakpoint group. Used for display. */
     protected String groupName;
     /** True if this breakpoint group is enabled. */
@@ -61,8 +61,6 @@ public class BreakpointGroup implements Serializable {
     protected List breakpointList;
     /** The breakpoint group to which we belong (always non-null). */
     protected BreakpointGroup parentGroup;
-    /** serial version */
-    static final long serialVersionUID = -4595947731582413285L;
 
     /**
      * Creates a BreakpointGroup with the default parameters.
@@ -113,7 +111,7 @@ public class BreakpointGroup implements Serializable {
      *
      * @return  number of breakpoints in this group and its subgroups.
      * @see #breakpointCount(boolean)
-     */ 
+     */
     public int breakpointCount() {
         // By default, get our count plus counts from subgroups.
         return breakpointCount(true);
@@ -126,7 +124,7 @@ public class BreakpointGroup implements Serializable {
      *                  this group's subgroups.
      * @return  number of breakpoints in this group.
      * @see #breakpointCount()
-     */ 
+     */
     public int breakpointCount(boolean recurse) {
         // First get our own count of breakpoints.
         int count = breakpointList.size();
@@ -198,7 +196,7 @@ public class BreakpointGroup implements Serializable {
      * the counts from this group's subgroups.
      *
      * @return  number of breakpoint groups in this group.
-     */ 
+     */
     public int groupCount() {
         return groupCount(true);
     } // groupCount
@@ -210,7 +208,7 @@ public class BreakpointGroup implements Serializable {
      *                  this group's subgroups.
      * @return  number of groups in this group.
      * @see #groupCount()
-     */ 
+     */
     public int groupCount(boolean recurse) {
         // First get our own count of groups.
         int count = groupList.size();
@@ -266,6 +264,113 @@ public class BreakpointGroup implements Serializable {
         }
     } // isEnabled
 
+    /**
+     * Reads the breakpoint properties from the given preferences node.
+     *
+     * @param  prefs  Preferences node from which to initialize this
+     *                breakpoint.
+     * @return  true if successful, false otherwise.
+     */
+    public boolean readObject(Preferences prefs) {
+        isEnabled = prefs.getBoolean("isEnabled", true);
+        groupName = prefs.get("groupName", "(undefined)");
+
+        // Look for persisted breakpoints.
+        Session session = getSession();
+        try {
+            int n = 1;
+            String subname = "breakpoint" + n;
+            while (prefs.nodeExists(subname)) {
+                // We have breakpoints.
+                Preferences subprefs = prefs.node(subname);
+                String cname = subprefs.get("class", null);
+                if (cname != null) {
+                    Class bclass = Class.forName(cname);
+                    Breakpoint bp = (Breakpoint) bclass.newInstance();
+                    addBreakpoint(bp);
+                    if (!bp.readObject(subprefs)) {
+                        String msg = MessageFormat.format(
+                            Bundle.getString("error.readBreakpoint"),
+                            new Object[] { String.valueOf(n) });
+                        session.getUIAdapter().showMessage(
+                            UIAdapter.MESSAGE_WARNING, msg);
+                        removeBreakpoint(bp);
+                    } else {
+                        bp.init();
+                    }
+                } else {
+                    String msg = MessageFormat.format(
+                        Bundle.getString("error.breakpointNoClass"),
+                        new Object[] { String.valueOf(n) });
+                    session.getUIAdapter().showMessage(
+                        UIAdapter.MESSAGE_WARNING, msg);
+                }
+                n++;
+                subname = "breakpoint" + n;
+            }
+
+        } catch (ClassNotFoundException cnfe) {
+            String msg = MessageFormat.format(
+                Bundle.getString("error.readBreakpoint"),
+                new Object[] { cnfe });
+            session.getUIAdapter().showMessage(
+                UIAdapter.MESSAGE_WARNING, msg);
+            return false;
+        } catch (IllegalAccessException iae) {
+            String msg = MessageFormat.format(
+                Bundle.getString("error.readBreakpoint"),
+                new Object[] { iae });
+            session.getUIAdapter().showMessage(
+                UIAdapter.MESSAGE_WARNING, msg);
+            return false;
+        } catch (InstantiationException ie) {
+            String msg = MessageFormat.format(
+                Bundle.getString("error.readBreakpoint"),
+                new Object[] { ie });
+            session.getUIAdapter().showMessage(
+                UIAdapter.MESSAGE_WARNING, msg);
+            return false;
+        } catch (BackingStoreException bse) {
+            String msg = MessageFormat.format(
+                Bundle.getString("error.readBreakpoint"),
+                new Object[] { bse });
+            session.getUIAdapter().showMessage(
+                UIAdapter.MESSAGE_WARNING, msg);
+            return false;
+        }
+
+        // Look for persisted breakpoint groups.
+        try {
+            int n = 1;
+            String subname = "group" + n;
+            while (prefs.nodeExists(subname)) {
+                // We have breakpoints.
+                Preferences subprefs = prefs.node(subname);
+                BreakpointGroup bg = new BreakpointGroup();
+                addBreakpointGroup(bg);
+                if (!bg.readObject(subprefs)) {
+                    String msg = MessageFormat.format(
+                        Bundle.getString("error.readBreakpointGroup"),
+                        new Object[] { String.valueOf(n) });
+                    session.getUIAdapter().showMessage(
+                        UIAdapter.MESSAGE_WARNING, msg);
+                    removeBreakpointGroup(bg);
+                }
+                n++;
+                subname = "group" + n;
+            }
+
+        } catch (BackingStoreException bse) {
+            String msg = MessageFormat.format(
+                Bundle.getString("error.readBreakpointGroup"),
+                new Object[] { bse });
+            session.getUIAdapter().showMessage(
+                UIAdapter.MESSAGE_WARNING, msg);
+            return false;
+        }
+        return true;
+    } // readObject
+
     /**
      * Removes the given breakpoint from this breakpoint group.
      *
@@ -328,6 +433,64 @@ public class BreakpointGroup implements Serializable {
         parentGroup = parent;
     } // setParent
 
+    /**
+     * Returns the string representation of this group.
+     *
+     * @return  string of this.
+     */
+    public String toString() {
+        String name = getName();
+        return name == null ? "<no-name>" : name;
+    } // toString
+
+    /**
+     * Writes the breakpoint properties to the given preferences node.
+     * It is assumed that the preferences node is completely empty.
+     *
+     * @param  prefs  Preferences node to which to serialize this
+     *                breakpoint.
+     * @return  true if successful, false otherwise.
+     */
+    public boolean writeObject(Preferences prefs) {
+        prefs.putBoolean("isEnabled", isEnabled);
+        prefs.put("groupName", groupName);
+
+        // For each breakpoint, make a new preference node and
+        // pass that node to the breakpoint.
+        int n = 1;
+        Iterator iter = breakpointList.iterator();
+        Session session = getSession();
+        while (iter.hasNext()) {
+            Breakpoint bp = (Breakpoint) iter.next();
+            if (!bp.writeObject(prefs.node("breakpoint" + n))) {
+                String msg = MessageFormat.format(
+                    Bundle.getString("error.writeBreakpoint"),
+                    new Object[] { String.valueOf(bp.getNumber()) });
+                session.getUIAdapter().showMessage(
+                    UIAdapter.MESSAGE_WARNING, msg);
+            }
+            n++;
+        }
+
+        // For each subgroup, make a new preference node and
+        // pass that node to the subgroup.
+        n = 1;
+        iter = groupList.iterator();
+        while (iter.hasNext()) {
+            BreakpointGroup bg = (BreakpointGroup) iter.next();
+            if (!bg.writeObject(prefs.node("group" + n))) {
+                String msg = MessageFormat.format(
+                    Bundle.getString("error.writeBreakpointGroup"),
+                    new Object[] { bg.getName() });
+                session.getUIAdapter().showMessage(
+                    UIAdapter.MESSAGE_WARNING, msg);
+            }
+            n++;
+        }
+
+        return true;
+    } // writeObject
+
     /**
      * Iterator of Breakpoints. Supports recursing through breakpoint
      * groups. It traverses the breakpoint groups in a breadth-first
@@ -393,8 +556,6 @@ public class BreakpointGroup implements Serializable {
          * Returns the next element in the interation.
          *
          * @return  the next element in the iteration.
-         * @exception  NoSuchElementException
-         *             iteration has no more elements.
          */
         public Object next() {
             if (breakpointIterator.hasNext()) {
@@ -446,6 +607,8 @@ public class BreakpointGroup implements Serializable {
         /**
          * Puts all of the current <code>group</code>'s groups onto
          * the group queue.
+         *
+         * @param  group  breakpoint group.
          */
         protected void fillGroupQueue(BreakpointGroup group) {
             Iterator groups = group.groups();
@@ -469,8 +632,6 @@ public class BreakpointGroup implements Serializable {
          * Returns the next element in the interation.
          *
          * @return  the next element in the iteration.
-         * @exception  NoSuchElementException
-         *             iteration has no more elements.
          */
         public Object next() {
             // Search for more groups.
