@@ -1,6 +1,6 @@
 /*********************************************************************
  *
- *      Copyright (C) 1999-2002 Nathan Fiedler
+ *      Copyright (C) 1999-2004 Nathan Fiedler
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -16,53 +16,58 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
- * PROJECT:     JSwat
- * MODULE:      Panel
- * FILE:        BreakPanel.java
- *
- * AUTHOR:      Nathan Fiedler
- *
- * REVISION HISTORY:
- *      Name    Date            Description
- *      ----    ----            -----------
- *      nf      04/28/99        Initial version
- *      nf      10/07/99        Upgraded to support parallel changes
- *      nf      06/10/01        Disembowled this class; rewrite someday.
- *      nf      07/09/01        Rewrote and added nice features.
- *      nf      09/01/01        Implemented request #181;
- *                              show breakpoint source using 'sourceFile'.
- *      nf      09/10/01        Implemented request #234
- *      tr      12/09/01        Error message if unable to map breakpoint
- *                              to a source file.
- *
- * DESCRIPTION:
- *      Defines the class responsible for displaying a list of
- *      user-defined breakpoints.
- *
- * $Id: BreakPanel.java 639 2002-10-27 22:08:27Z nfiedler $
+ * $Id: BreakPanel.java 1157 2004-02-25 09:28:54Z nfiedler $
  *
  ********************************************************************/
 
 package com.bluemarsh.jswat.panel;
 
-import com.bluemarsh.jswat.JSwat;
 import com.bluemarsh.jswat.PathManager;
 import com.bluemarsh.jswat.Session;
 import com.bluemarsh.jswat.SourceSource;
 import com.bluemarsh.jswat.action.ActionTable;
-import com.bluemarsh.jswat.breakpoint.*;
+import com.bluemarsh.jswat.breakpoint.Breakpoint;
+import com.bluemarsh.jswat.breakpoint.BreakpointEvent;
+import com.bluemarsh.jswat.breakpoint.BreakpointGroup;
+import com.bluemarsh.jswat.breakpoint.BreakpointListener;
+import com.bluemarsh.jswat.breakpoint.BreakpointManager;
+import com.bluemarsh.jswat.breakpoint.GroupEvent;
+import com.bluemarsh.jswat.breakpoint.GroupListener;
+import com.bluemarsh.jswat.breakpoint.LocatableBreakpoint;
 import com.bluemarsh.jswat.breakpoint.ui.EditorDialog;
+import com.bluemarsh.jswat.event.SessionEvent;
 import com.bluemarsh.jswat.ui.SessionFrameMapper;
 import com.bluemarsh.jswat.ui.UIAdapter;
-import com.sun.jdi.Location;
-import java.awt.*;
-import java.awt.event.*;
+import com.sun.jdi.ReferenceType;
+import java.awt.Component;
+import java.awt.Font;
+import java.awt.Frame;
+import java.awt.GridBagConstraints;
+import java.awt.GridBagLayout;
+import java.awt.event.ActionEvent;
+import java.awt.event.ActionListener;
+import java.awt.event.MouseEvent;
+import java.awt.event.MouseListener;
+import java.awt.event.WindowAdapter;
+import java.awt.event.WindowEvent;
+import java.beans.PropertyChangeEvent;
+import java.beans.PropertyChangeListener;
 import java.io.File;
 import java.io.IOException;
-import java.net.URL;
+import java.text.MessageFormat;
+import java.util.ConcurrentModificationException;
 import java.util.Iterator;
-import javax.swing.*;
-import javax.swing.tree.*;
+import javax.swing.JButton;
+import javax.swing.JComponent;
+import javax.swing.JPanel;
+import javax.swing.JScrollPane;
+import javax.swing.JTree;
+import javax.swing.UIManager;
+import javax.swing.tree.DefaultMutableTreeNode;
+import javax.swing.tree.DefaultTreeCellRenderer;
+import javax.swing.tree.DefaultTreeModel;
+import javax.swing.tree.MutableTreeNode;
+import javax.swing.tree.TreePath;
 
 /**
  * Class BreakPanel is responsible for displaying all known breakpoints
@@ -74,82 +79,35 @@ import javax.swing.tree.*;
  *
  * @author  Nathan Fiedler
  */
-public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupListener {
+public class BreakPanel extends AbstractPanel
+    implements BreakpointListener, GroupListener, MouseListener,
+               PropertyChangeListener {
     /** Tree representing the breakoint groups and breakpoints. */
-    protected JTree uiTree;
+    private JTree uiTree;
     /** Containing component object. */
-    protected JComponent uiComponent;
-    /** Session that we are listening to. */
-    protected Session owningSession;
+    private JComponent uiComponent;
 
     /**
      * Constructs a BreakPanel.
      */
     public BreakPanel() {
         // Don't try to build out anything just now. Wait for the
-        // call to the init() method.
+        // call to the opened() method.
         BasicTreeNode root = new BasicTreeNode("junk");
         uiTree = new JTree(new DefaultTreeModel(root));
-        uiTree.setCellRenderer(new DefaultTreeCellRenderer() {
-                public Component getTreeCellRendererComponent
-                    (JTree tree, Object value,
-                     boolean isSelected, boolean isExpanded,
-                     boolean leaf, int row, boolean hasFocus) {
-                    // Get the default renderer first.
-                    Component renderer = super.getTreeCellRendererComponent
-                        (tree, value, isSelected, isExpanded,
-                         leaf, row, hasFocus);
-                    BasicTreeNode node = (BasicTreeNode) value;
-                    setIcon(node.getIcon());
-                    return renderer;
-                }
-            });
-
-        // Add a mouse listener to show breakpoint properties.
-        uiTree.addMouseListener(new MouseAdapter() {
-                public void mouseClicked(MouseEvent e) {
-                    if (e.getClickCount() == 2) {
-                        TreePath selPath = uiTree.getPathForLocation
-                            (e.getX(), e.getY());
-                        if (selPath != null) {
-                            Object[] path = selPath.getPath();
-                            if (path.length > 1) {
-                                BasicTreeNode node = (BasicTreeNode)
-                                    selPath.getLastPathComponent();
-                                if (node.isLeaf()) {
-                                    // See if it is a breakpoint.
-                                    if (node instanceof BreakpointNode) {
-                                        Breakpoint bp = (Breakpoint)
-                                            node.getUserObject();
-                                        editBreakpoint(bp);
-                                    }
-                                }
-                            }
-                        }
-                    }
-                }
-            });
-
+        uiTree.setShowsRootHandles(true);
+        uiTree.setCellRenderer(new BreakpointRenderer());
+        uiTree.addMouseListener(this);
         uiComponent = buildUI(new JScrollPane(uiTree));
     } // BreakPanel
 
-    /**
-     * Called when the Session is about to begin an active debugging
-     * session. That is, JSwat is about to debug a debuggee VM.
-     * Panels are not activated in any particular order.
-     *
-     * @param  session  Session being activated.
-     */
-    public void activate(Session session) {
-    } // activate
-
     /**
      * Invoked when a breakpoint has been added.
      *
      * @param  e  breakpoint change event
      */
     public void breakpointAdded(BreakpointEvent e) {
-        refresh(owningSession);
+        refreshLater();
     } // breakpointAdded
 
     /**
@@ -158,7 +116,7 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
      * @param  e  breakpoint change event
      */
     public void breakpointModified(BreakpointEvent e) {
-        refresh(owningSession);
+        refreshLater();
     } // breakpointModified
 
     /**
@@ -169,11 +127,7 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
     public void breakpointRemoved(BreakpointEvent e) {
         // We're notified before the breakpoint is actually removed,
         // so call the refresh() method later.
-        SwingUtilities.invokeLater(new Runnable() {
-                public void run() {
-                    refresh(owningSession);
-                }
-            });
+        refreshLater();
     } // breakpointRemoved
 
     /**
@@ -196,14 +150,11 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
         newComp.add(tree);
 
         // Reset the constraints.
-        gbc.fill = GridBagConstraints.NONE;
+        gbc.fill = GridBagConstraints.HORIZONTAL;
         gbc.gridheight = 1;
         gbc.gridy = 0;
         gbc.weightx = 0.0;
 
-        // Buttons are spaced at least 2 pixels apart.
-        gbc.insets = new Insets(1, 1, 1, 1);
-
         JButton button;
 
         // Add a button to edit the properties of the selected breakpoint.
@@ -263,13 +214,21 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
 
         // Add a button to allow adding breakpoints.
         button = new JButton(Bundle.getString("Break.addLabel"));
-        button.addActionListener(ActionTable.getAction("setBreak"));
+        try {
+            button.addActionListener(ActionTable.getAction("setBreak"));
+        } catch (Exception e) {
+            button.setEnabled(false);
+        }
         gbl.setConstraints(button, gbc);
         newComp.add(button);
 
         // Add a button to allow adding breakpoint groups.
         button = new JButton(Bundle.getString("Break.addGroupLabel"));
-        button.addActionListener(ActionTable.getAction("addBreakGroup"));
+        try {
+            button.addActionListener(ActionTable.getAction("addBreakGroup"));
+        } catch (Exception e) {
+            button.setEnabled(false);
+        }
         gbl.setConstraints(button, gbc);
         newComp.add(button);
 
@@ -279,8 +238,8 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
     /**
      * Fill in the given tree node with the breakpoint's contents.
      *
-     * @param  brkpgrp  breakpoint group to build out.
-     * @param  node     tree node to fill in.
+     * @param  brkgrp  breakpoint group to build out.
+     * @param  node    tree node to fill in.
      */
     protected void buildGroupNode(BreakpointGroup brkgrp,
                                   MutableTreeNode node) {
@@ -305,28 +264,18 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
     } // buildGroupNode
 
     /**
-     * Called when the Session is closing down this panel, generally
-     * just after the panel has been removed from the Session.
+     * Called when the Session is about to be closed.
      *
-     * @param  session  Session closing the panel.
+     * @param  sevt  session event.
      */
-    public void close(Session session) {
+    public void closing(SessionEvent sevt) {
+        UIManager.removePropertyChangeListener(this);
         BreakpointManager brkman = (BreakpointManager)
-            session.getManager(BreakpointManager.class);
+            sevt.getSession().getManager(BreakpointManager.class);
         brkman.removeBreakListener(this);
         brkman.removeGroupListener(this);
-    } // close
-
-    /**
-     * Called when the Session is about to end an active debugging
-     * session. That is, JSwat is about to terminate the connection
-     * with the debuggee VM.
-     * Panels are not deactivated in any particular order.
-     *
-     * @param  session  Session being deactivated.
-     */
-    public void deactivate(Session session) {
-    } // deactivate
+        super.closing(sevt);
+    } // closing
 
     /**
      * Deletes the currently selected breakpoints and groups.
@@ -386,12 +335,12 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
         editor.addWindowListener(new WindowAdapter() {
                 // For some reason, windowClosing() isn't sent.
                 public void windowClosed(WindowEvent we) {
-                                // When the editor dialog closes, refresh
-                                // the breakpoints panel.
-                    refresh(owningSession);
+                    // When the editor dialog closes, refresh
+                    // the breakpoints panel.
+                    refreshLater();
                 }
             });
-        editor.show();
+        editor.setVisible(true);
     } // editBreakpoint
 
     /**
@@ -451,7 +400,7 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
      * @param  event  group change event
      */
     public void groupAdded(GroupEvent event) {
-        refresh(owningSession);
+        refreshLater();
     } // groupAdded
 
     /**
@@ -460,7 +409,7 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
      * @param  event  group change event
      */
     public void groupDisabled(GroupEvent event) {
-        refresh(owningSession);
+        refreshLater();
     } // groupDisabled
 
     /**
@@ -469,7 +418,7 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
      * @param  event  group change event
      */
     public void groupEnabled(GroupEvent event) {
-        refresh(owningSession);
+        refreshLater();
     } // groupEnabled
 
     /**
@@ -478,24 +427,77 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
      * @param  event  group change event
      */
     public void groupRemoved(GroupEvent event) {
-        refresh(owningSession);
+        refreshLater();
     } // groupRemoved
 
     /**
-     * Called when the Session is ready to initialize this panel,
-     * generally just after the panel has been added to the Session.
+     * Invoked when the mouse button has been clicked (pressed and
+     * released) on a component.
      *
-     * @param  session  Session initializing this panel.
+     * @param  e  mouse event.
      */
-    public void init(Session session) {
-        owningSession = session;
-        // Need to build out the tree as soon as possible.
-        refresh(session);
+    public void mouseClicked(MouseEvent e) {
+        if (e.getClickCount() == 2) {
+            TreePath selPath = uiTree.getPathForLocation(e.getX(), e.getY());
+            if (selPath != null) {
+                Object[] path = selPath.getPath();
+                if (path.length > 1) {
+                    BasicTreeNode node = (BasicTreeNode)
+                        selPath.getLastPathComponent();
+                    if (node.isLeaf()) {
+                        // See if it is a breakpoint.
+                        if (node instanceof BreakpointNode) {
+                            Breakpoint bp = (Breakpoint) node.getUserObject();
+                            editBreakpoint(bp);
+                        }
+                    }
+                }
+            }
+        }
+    } // mouseClicked
+
+    /**
+     * Invoked when a mouse button has been pressed on a component.
+     *
+     * @param  e  mouse event.
+     */
+    public void mousePressed(MouseEvent e) { }
+
+    /**
+     * Invoked when a mouse button has been released on a component.
+     *
+     * @param  e  mouse event.
+     */
+    public void mouseReleased(MouseEvent e) { }
+
+    /**
+     * Invoked when the mouse enters a component.
+     *
+     * @param  e  mouse event.
+     */
+    public void mouseEntered(MouseEvent e) { }
+
+    /**
+     * Invoked when the mouse exits a component.
+     *
+     * @param  e  mouse event.
+     */
+    public void mouseExited(MouseEvent e) { }
+
+    /**
+     * Called after the Session has added this listener to the Session
+     * listener list.
+     *
+     * @param  session  the Session.
+     */
+    public void opened(Session session) {
+        super.opened(session);
         BreakpointManager brkman = (BreakpointManager)
             session.getManager(BreakpointManager.class);
         brkman.addBreakListener(this);
         brkman.addGroupListener(this);
-    } // init
+        UIManager.addPropertyChangeListener(this);
+    } // opened
 
     /**
      * Perform some action on the set of selected breakpoints
@@ -532,13 +534,29 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
 
         // Rebuild the tree if anything changed.
         if (rebuild) {
-            refresh(owningSession);
+            refreshLater();
         }
     } // processSelections
 
     /**
-     * Update the display on the screen. Use the given Session
-     * to fetch the desired data.
+     * This method gets called when a bound property is changed.
+     *
+     * @param  evt  a PropertyChangeEvent object describing the event source
+     *              and the property that has changed.
+     */
+    public void propertyChange(PropertyChangeEvent evt) {
+        if (evt.getPropertyName().equals("lookAndFeel")) {
+            Object renderer = uiTree.getCellRenderer();
+            if (renderer instanceof JComponent) {
+                ((JComponent) renderer).updateUI();
+            }
+        }
+    } // propertyChange
+
+    /**
+     * Update the display on the screen. Use the given Session to fetch
+     * the desired data. This must be run on the AWT event dispatching
+     * thread.
      *
      * @param  session  debugging Session object.
      */
@@ -552,8 +570,14 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
         GroupNode groupNode = new GroupNode(group);
         DefaultTreeModel model = new DefaultTreeModel(groupNode);
 
-        // Recursively build out the rest of the tree.
-        buildGroupNode(group, groupNode);
+        try {
+            // Recursively build out the rest of the tree.
+            buildGroupNode(group, groupNode);
+        } catch (ConcurrentModificationException cme) {
+            // Whoops, something changed in the breakpoints.
+            // Let's bail and hope we get called again later.
+            return;
+        }
 
         // Set the model now that it has been built out.
         uiTree.setModel(model);
@@ -576,62 +600,138 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
             return;
         }
 
-        // Try the way that is most likely to succeed first.
+        // First try the way that is most likely to succeed.
         LocatableBreakpoint lbp = (LocatableBreakpoint) bp;
         PathManager pathman = (PathManager)
             owningSession.getManager(PathManager.class);
-        // Won't work if the class name is pattern.
         SourceSource src = null;
+        // This won't work if the class name is wild-carded.
         try {
             src = pathman.mapSource(lbp.getClassName());
         } catch (IOException ioe) {
-            ioe.printStackTrace();
+            // This case is handled below.
         }
-        int line = lbp.getLineNumber();
 
-        if (line < 0) {
+        if (src == null) {
+            // See if the source name and package are available.
+            StringBuffer filepath = new StringBuffer();
+            String path = lbp.getPackageName();
+            if (path != null) {
+                filepath.append(path.replace('.', File.separatorChar));
+                filepath.append(File.separatorChar);
+            }
+            String name = lbp.getSourceName();
+            if (name != null) {
+                filepath.append(name);
+            }
+            if (filepath.length() > 0) {
+                src = pathman.mapFile(filepath.toString());
+            }
+        }
+        if (src == null) {
             // That didn't work, try the resolved location.
-            Location loc = lbp.getLocation();
-            if (loc == null) {
-                JOptionPane.showMessageDialog(
-                    null,
-                    Bundle.getString("Break.bpHasNoLocation"),
-                    Bundle.getString("Break.errorTitle"),
-                    JOptionPane.ERROR_MESSAGE);
+            ReferenceType clazz = lbp.getReferenceType();
+            if (clazz == null) {
+                owningSession.getUIAdapter().showMessage(
+                    UIAdapter.MESSAGE_ERROR,
+                    Bundle.getString("Break.bpHasNoLocation"));
                 return;
             }
             try {
-                src = pathman.mapSource(loc.declaringType());
+                src = pathman.mapSource(clazz);
             } catch (IOException ioe) {
-                ioe.printStackTrace();
+                // This case is handled below.
             }
-            line = loc.lineNumber();
         }
 
-        if ((src != null) && (line > 0)) {
+        int line = lbp.getLineNumber();
+        if (src != null && src.exists() && line > 0) {
             // Open a source view to that line.
             UIAdapter adapter = owningSession.getUIAdapter();
             if (!adapter.showFile(src, line, 0)) {
-                owningSession.getStatusLog().writeln(
-                    Bundle.getString("Break.openFileFailed"));
+                String msg = MessageFormat.format(
+                    Bundle.getString("couldntOpenFileMsg"),
+                    new Object[] { src.getName() });
+                owningSession.getUIAdapter().showMessage(
+                    UIAdapter.MESSAGE_ERROR, msg);
             }
         } else {
-            owningSession.getStatusLog().writeln(
-                Bundle.getString("Break.showFailed1") + " " +
-                lbp.getClassName() +
-                "\n" + Bundle.getString("Break.showFailed2"));
+            String msg = MessageFormat.format(
+                Bundle.getString("Break.showFailed"),
+                new Object[] { lbp.getClassName() });
+            owningSession.getUIAdapter().showMessage(
+                UIAdapter.MESSAGE_ERROR, msg);
         }
     } // showSource
 
+    /**
+     * Class BreakpointRenderer implements a tree cell renderer that
+     * draws breakpoints in the breakpoint panel.
+     */
+    protected class BreakpointRenderer extends DefaultTreeCellRenderer {
+
+        /**
+         * Constructs a BreakpointRenderer.
+         */
+        public BreakpointRenderer() {
+            super();
+            // Kill all of those nasty icons.
+            setClosedIcon(null);
+            setLeafIcon(null);
+            setOpenIcon(null);
+        } // BreakpointRenderer
+
+        /**
+         * Return the component for rendering the identified cell.
+         *
+         * @param  tree        the tree.
+         * @param  value       the node user object.
+         * @param  isSelected  true if node is selected.
+         * @param  isExpanded  true if node is expanded.
+         * @param  leaf        true if node is a leaf.
+         * @param  row         row of the tree.
+         * @param  hasFocus    true if node has focus.
+         * @return  rendering component.
+         */
+        public Component getTreeCellRendererComponent(
+            JTree tree, Object value, boolean isSelected, boolean isExpanded,
+            boolean leaf, int row, boolean hasFocus) {
+
+            super.getTreeCellRendererComponent(
+                tree, value, isSelected, isExpanded, leaf, row, hasFocus);
+            BasicTreeNode node = (BasicTreeNode) value;
+            Font font = getFont();
+            if (!font.isPlain()) {
+                font = font.deriveFont(Font.PLAIN);
+            }
+            setFont(node.getFont(font));
+            return this;
+        } // getTreeCellRendererComponent
+
+        /**
+         * Called when the look and feel is changing.
+         */
+        public void updateUI() {
+            super.updateUI();
+            // Hmm, why doesn't it do this automatically?
+            setTextSelectionColor(UIManager.getColor(
+                                      "Tree.selectionForeground"));
+            setTextNonSelectionColor(UIManager.getColor(
+                                         "Tree.textForeground"));
+            setBackgroundSelectionColor(UIManager.getColor(
+                                            "Tree.selectionBackground"));
+            setBackgroundNonSelectionColor(UIManager.getColor(
+                                               "Tree.textBackground"));
+            setBorderSelectionColor(UIManager.getColor(
+                                        "Tree.selectionBorderColor"));
+        } // updateUI
+    } // BreakpointRenderer
+
     /**
      * Class BasicTreeNode is the base class for both breakpoint
      * nodes and breakpoint group nodes.
      */
     protected class BasicTreeNode extends DefaultMutableTreeNode {
-        /** Icon for disabled breakpoints. */
-        protected Icon disabledIcon;
-        /** Icon for resolved breakpoints. */
-        protected Icon resolvedIcon;
 
         /**
          * Constructs a BasicTreeNode.
@@ -653,38 +753,16 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
         } // BasicTreeNode
 
         /**
-         * Returns the icon appropriate for this node.
+         * Returns the font appropriate for this node, based on the
+         * given font.
          *
-         * @return  icon appropriate for this node.
+         * @param  font  font from which to derive a new one; the initial
+         *               style is plain.
+         * @return  font appropriate for this node.
          */
-        public Icon getIcon() {
-            if (disabledIcon == null) {
-                JSwat swat = JSwat.instanceOf();
-                URL url = swat.getResource("brkDisabledImage");
-                disabledIcon = new ImageIcon(url);
-                url = swat.getResource("brkResolvedImage");
-                resolvedIcon = new ImageIcon(url);
-            }
-
-            Object o = getUserObject();
-            if (o instanceof Breakpoint) {
-                Breakpoint bp = (Breakpoint) o;
-                if (!bp.isEnabled()) {
-                    // Breakpoint is disabled.
-                    return disabledIcon;
-                } if (bp.isResolved()) {
-                    // Breakpoint is resolved.
-                    return resolvedIcon;
-                }
-            } if (o instanceof BreakpointGroup) {
-                BreakpointGroup grp = (BreakpointGroup) o;
-                if (!grp.isEnabled()) {
-                    // Breakpoint group is disabled.
-                    return disabledIcon;
-                }
-            }
+        public Font getFont(Font font) {
             return null;
-        } // getIcon
+        } // getFont
     } // BasicTreeNode
 
     /**
@@ -701,6 +779,25 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
             super(brk, false);
         } // BreakpointNode
 
+        /**
+         * Returns the font appropriate for this node, based on the
+         * given font.
+         *
+         * @param  font  font from which to derive a new one; the initial
+         *               style is plain.
+         * @return  font appropriate for this node.
+         */
+        public Font getFont(Font font) {
+            Breakpoint bp = (Breakpoint) getUserObject();
+            if (!bp.isEnabled()) {
+                font = font.deriveFont(Font.ITALIC);
+            }
+            if (bp.isResolved()) {
+                font = font.deriveFont(Font.BOLD | font.getStyle());
+            }
+            return font;
+        } // getFont
+
         /**
          * Returns the name of the breakpoint group.
          *
@@ -727,6 +824,22 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
             super(brkgrp);
         } // GroupNode
 
+        /**
+         * Returns the font appropriate for this node, based on the
+         * given font.
+         *
+         * @param  font  font from which to derive a new one; the initial
+         *               style is plain.
+         * @return  font appropriate for this node.
+         */
+        public Font getFont(Font font) {
+            BreakpointGroup grp = (BreakpointGroup) getUserObject();
+            if (!grp.isEnabled()) {
+                font = font.deriveFont(Font.ITALIC);
+            }
+            return font;
+        } // getFont
+
         /**
          * Returns the name of the breakpoint group.
          *
@@ -751,6 +864,6 @@ public class BreakPanel extends JSwatPanel implements BreakpointListener, GroupL
          * @param  o  user object from a node of the tree.
          * @return  true if the tree should be rebuilt.
          */
-        public boolean processNode(BreakpointManager brkman, Object o);
+        boolean processNode(BreakpointManager brkman, Object o);
     } // NodeProcessor
 } // BreakPanel
