@@ -1,6 +1,6 @@
 /*********************************************************************
  *
- *      Copyright (C) 2001-2002 Nathan Fiedler
+ *      Copyright (C) 2001-2003 Nathan Fiedler
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -29,65 +29,69 @@
  *      nf      01/06/02        Fixed bug 377
  *      nf      02/15/02        Fixed bug 393
  *      nf      02/20/02        Fixed bug 399
- *      nf      03/28/02        Fixed bug 414 (contributed by Masaru O.)
- *      nf      04/15/02        Implemented 433 (patch from Masaru O.)
+ *      nf      03/28/02        Fixed bug 414
+ *      nf      04/15/02        Implemented RFE 433
+ *      nf      10/20/02        Fixed bug 650
+ *      nf      12/28/02        Fixed bug 693
+ *      nf      01/26/03        Removed logging calls
+ *      nf      11/05/03        Implemented RFE 658
  *
- * DESCRIPTION:
- *      This file contains the SourceViewTextArea class definition.
- *
- * $Id: SourceViewTextArea.java 629 2002-10-26 23:03:26Z nfiedler $
+ * $Id: SourceViewTextArea.java 1051 2003-11-17 05:51:45Z nfiedler $
  *
  ********************************************************************/
 
 package com.bluemarsh.jswat.view;
 
-import com.bluemarsh.adt.PriorityList;
-import com.bluemarsh.jswat.Session;
-import com.bluemarsh.jswat.ContextManager;
-import com.bluemarsh.jswat.FieldNotObjectException;
-import com.bluemarsh.jswat.report.Category;
-import com.bluemarsh.jswat.ui.SessionFrameMapper;
-import com.bluemarsh.jswat.util.VariableUtils;
-import com.sun.jdi.*;
-import java.awt.*;
-import java.awt.event.*;
-import java.io.*;
+import com.bluemarsh.jswat.Defaults;
+import com.bluemarsh.jswat.util.PriorityList;
+import java.awt.Color;
+import java.awt.Dimension;
+import java.awt.FontMetrics;
+import java.awt.Graphics;
+import java.awt.Point;
+import java.awt.Rectangle;
+import java.awt.event.MouseEvent;
+import java.awt.event.MouseListener;
+import java.awt.event.MouseMotionListener;
 import java.util.ArrayList;
-import javax.swing.*;
-import javax.swing.text.*;
+import java.util.Iterator;
+import javax.swing.JComponent;
+import javax.swing.JViewport;
+import javax.swing.Scrollable;
+import javax.swing.SwingConstants;
+import javax.swing.ToolTipManager;
+import javax.swing.text.BadLocationException;
+import java.util.prefs.Preferences;
 
 /**
- * Class SourceViewTextArea is a quasi-text component that implements
- * the SourceViewArea interface for easily getting the line number
- * that corresponds to a view coordinate.
+ * Class SourceViewTextArea is a read-only text component.
  *
- * <p>In addition, this class has the ability to interpret the value
- * of the string under the mouse pointer. That is, it will read the
- * string and assume it is a variable reference. If it is a variable,
- * the value is displayed in the form of a tooltip.</p>
+ * <p>In addition, this class has the ability to interpret the value of the
+ * string under the mouse pointer. That is, it will read the string and
+ * assume it is a variable reference. If it is a variable, the value is
+ * displayed in the form of a tooltip.</p>
  *
  * @author  Nathan Fiedler
  * @author  David Taylor
+ * @author  Masaru Ohba
  */
-public class SourceViewTextArea extends JComponent implements Scrollable, SourceViewArea {
-    /** If true, does additional debugging output to console. */
-    protected static final boolean DEBUG = false;
-    /** Debug reporting category. */
-    protected static Category logCategory = Category.instanceOf("sourceview");
+public class SourceViewTextArea extends JComponent implements Scrollable {
     /** Color for the gutter background. */
-    protected static Color gutterColor = Color.lightGray;
+    private static Color gutterColor = Color.lightGray;
     /** Content that contains the text. */
-    protected SourceContent content;
+    private SourceContent content;
     /** Draw layer for drawing the current text selection. */
-    protected SelectionDrawLayer selectionDrawLayer;
+    private SelectionDrawLayer selectionDrawLayer;
     /** Draw layers in priority order. */
-    protected PriorityList drawLayersList;
+    private PriorityList drawLayersList;
     /** Draw layers in priority order. */
-    protected PriorityList gutterLayersList;
-    /** Width of the gutter. */
-    protected int gutterWidth;
+    private PriorityList gutterLayersList;
+    /** Tooltip producers in priority order. */
+    private PriorityList tooltipProducersList;
+    /** Width of the view gutter. */
+    private int gutterWidth;
     /** Width of the source code. */
-    protected int sourceWidth;
+    private int sourceWidth;
 
     /**
      * Constructs a SourceViewTextArea for the given text.
@@ -99,14 +103,15 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         ToolTipManager.sharedInstance().registerComponent(this);
         drawLayersList = new PriorityList();
         gutterLayersList = new PriorityList();
+        tooltipProducersList = new PriorityList();
         selectionDrawLayer = new SelectionDrawLayer();
-        addDrawLayer(selectionDrawLayer, DrawLayer.PRIORITY_SELECTION);
+        addDrawLayer(selectionDrawLayer);
         MouseSelector selector = new MouseSelector();
         addMouseListener(selector);
         addMouseMotionListener(selector);
         setBackground(Color.white);
-        // This is only half of it; we still need to do the work
-        // in mouseDragged() method of listener.
+        // This is only half of it; we still need to do the work in the
+        // mouseDragged() method.
         setAutoscrolls(true);
     } // SourceViewTextArea
 
@@ -114,38 +119,47 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
      * Adds the given draw layer to this text area.
      *
      * @param  layer  draw layer to add.
-     * @param  prio   priority for the draw layer.
      */
-    public void addDrawLayer(DrawLayer layer, int prio) {
+    public void addDrawLayer(DrawLayer layer) {
         // Priorities are lower to higher in our world.
-        if (prio < DrawLayer.PRIORITY_HIGHEST ||
-            prio > DrawLayer.PRIORITY_LOWEST) {
+        int prio = layer.getPriority();
+        if (prio < DrawLayer.PRIORITY_HIGHEST
+            || prio > DrawLayer.PRIORITY_LOWEST) {
             throw new IllegalArgumentException("priority out of range");
         }
         drawLayersList.add(layer, prio);
-        if (DEBUG) {
-            System.out.println("draw layer added: " + layer.getClass());
-        }
     } // addDrawLayer
 
     /**
-     * Adds the given draw layer to this text area.
+     * Adds the given gutter draw layer to this text area.
      *
-     * @param  layer  draw layer to add.
-     * @param  prio   priority for the draw layer.
+     * @param  layer  gutter draw layer to add.
      */
-    public void addGutterLayer(GutterDrawLayer layer, int prio) {
+    public void addGutterLayer(GutterDrawLayer layer) {
         // Priorities are lower to higher in our world.
-        if (prio < GutterDrawLayer.PRIORITY_HIGHEST ||
-            prio > GutterDrawLayer.PRIORITY_LOWEST) {
+        int prio = layer.getPriority();
+        if (prio < GutterDrawLayer.PRIORITY_HIGHEST
+            || prio > GutterDrawLayer.PRIORITY_LOWEST) {
             throw new IllegalArgumentException("priority out of range");
         }
         gutterLayersList.add(layer, prio);
-        if (DEBUG) {
-            System.out.println("gutter layer added: " + layer.getClass());
-        }
     } // addGutterLayer
 
+    /**
+     * Adds the given tooltip producer to this text area.
+     *
+     * @param  prod  tooltip producer to add.
+     */
+    public void addTooltipProducer(TooltipProducer prod) {
+        // Priorities are lower to higher in our world.
+        int prio = prod.getPriority();
+        if (prio < TooltipProducer.PRIORITY_HIGHEST
+            || prio > TooltipProducer.PRIORITY_LOWEST) {
+            throw new IllegalArgumentException("priority out of range");
+        }
+        tooltipProducersList.add(prod, prio);
+    } // addTooltipProducer
+
     /**
      * Returns the source view text area content object.
      *
@@ -155,70 +169,99 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         return content;
     } // getContent
 
+    /**
+     * Calculate the width of the gutter and return the value in pixels.
+     *
+     * @return  gutter width in pixels.
+     */
+    protected int getGutterWidth() {
+        if (gutterWidth <= 0) {
+            return getGutterWidth(getFontMetrics(getFont()));
+        } else {
+            return gutterWidth;
+        }
+    } // getGutterWidth
+
+    /**
+     * Calculate the width of the gutter and return the value in pixels.
+     *
+     * @param  metrics  font metrics for calculating width.
+     * @return  gutter width in pixels.
+     */
+    protected int getGutterWidth(FontMetrics metrics) {
+        if (gutterWidth <= 0) {
+            Preferences prefs = Preferences.userRoot().node(
+                "com/bluemarsh/jswat/view");
+            if (prefs.getBoolean("viewLineNumbers",
+                    Defaults.VIEW_LINE_NUMBERS)) {
+                gutterWidth = metrics.stringWidth("999999");
+            } else {
+                // Give a little space for the breakpoint indicators.
+                gutterWidth = metrics.stringWidth("WW");
+            }
+        }
+        return gutterWidth;
+    } // getGutterWidth
+
     /**
      * Returns the word (if any) that contains position offset.
      *
      * @param  offset  the offset into the document.
      * @return  the word that contains offset, or null if no word is nearby.
+     * @throws  BadLocationException
+     *          if the text location is invalid.
      */
-    private String getIdentifierAt(int offset) {
+    public String getIdentifierAt(int offset)
+        throws BadLocationException {
         // This method could stand optimisation in terms or speed, and
         // also improvement to recognise x[i], etc.
-        try {
-            int line = content.getLineOfOffset(offset) + 1;
-            int lineEnd = content.getLineRealEndOffset(line);
-            char[] lineText = content.getBuffer();
+        int line = content.getLineOfOffset(offset) + 1;
+        int lineEnd = content.getLineEndOffset(line);
+        char[] lineText = content.getBuffer();
 
-            if (Character.isWhitespace(lineText[offset])) {
-                return null;
-            }
-
-            boolean foundMatch = false;
-            int endOfWord = offset;
-
-            // Find the end of the Java identifier.
-            for (int i = offset; i < lineEnd; i++) {
-                char ch = lineText[i];
-                if (!Character.isJavaIdentifierPart(ch) && ch != ']') {
-                    endOfWord = i;
-                    foundMatch = true;
-                    break;
-                }
-            }
-
-            if (!foundMatch) {
-                endOfWord = lineEnd - 1;
-            }
-
-            // Find the start of the Java identifier.
-            foundMatch = false;
-            int startOfWord = offset;
-            for (int i = offset; i >= 0; i--) {
-                char ch = lineText[i];
-                // Accept '.', '[', and ']' while going backwards so we
-                // build up a full variable name (e.g. "obj[i].field").
-                if (!Character.isJavaIdentifierPart(ch) &&
-                    ch != '.' && ch != '[' && ch != ']') {
-                    startOfWord = i + 1;
-                    foundMatch = true;
-                    break;
-                }
-            }
-
-            if (!foundMatch) {
-                startOfWord = 0;
-            }
-            if ((startOfWord < 0) || ((endOfWord - startOfWord + 1) < 1)) {
-                return null;
-            }
-            return new String(lineText, startOfWord, endOfWord - startOfWord);
-        } catch (BadLocationException ble) {
-            ble.printStackTrace();
-        } catch (StringIndexOutOfBoundsException sioobe) {
-            sioobe.printStackTrace();
+        if (Character.isWhitespace(lineText[offset])) {
+            return null;
         }
 
-        return null;
+        boolean foundMatch = false;
+        int endOfWord = offset;
+
+        // Find the end of the Java identifier.
+        for (int i = offset; i < lineEnd; i++) {
+            char ch = lineText[i];
+            if (!Character.isJavaIdentifierPart(ch) && ch != ']') {
+                endOfWord = i;
+                foundMatch = true;
+                break;
+            }
+        }
+
+        if (!foundMatch) {
+            endOfWord = lineEnd - 1;
+        }
+
+        // Find the start of the Java identifier.
+        foundMatch = false;
+        int startOfWord = offset;
+        for (int i = offset; i >= 0; i--) {
+            char ch = lineText[i];
+            // Accept '.', '[', and ']' while going backwards so we
+            // build up a full variable name (e.g. "obj[i].field").
+            if (!Character.isJavaIdentifierPart(ch)
+                && ch != '.' && ch != '[' && ch != ']') {
+                startOfWord = i + 1;
+                foundMatch = true;
+                break;
+            }
+        }
+
+        if (!foundMatch) {
+            startOfWord = 0;
+        }
+        if ((startOfWord < 0) || ((endOfWord - startOfWord + 1) < 1)) {
+            return null;
+        }
+        return new String(lineText, startOfWord, endOfWord - startOfWord);
     } // getIdentifierAt
 
     /**
@@ -249,9 +292,7 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         int height = content.getLineCount() * lineHeight;
         // Masaru Ohba's proportional font code.
         // width = gutterWidth + sourceWidth;
-        if (gutterWidth <= 0) {
-            gutterWidth = metrics.stringWidth("999999");
-        }
+        int gutterWidth = getGutterWidth(metrics);
         if (sourceWidth <= 0) {
             char[] buffer = content.getBuffer();
             int startOfLine = 0;
@@ -260,13 +301,13 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
             for (int line = 0 ; line < content.getLineCount() ; line++) {
                 try {
                     startOfLine = content.getLineStartOffset(line);
-                    endOfLine = content.getLineRealEndOffset(line);
+                    endOfLine = content.getLineEndOffset(line);
                 } catch (BadLocationException ble) {
                     startOfLine = 0;
                     endOfLine = 0;
                 }
                 lineWidth = metrics.charsWidth(buffer, startOfLine,
-                                               endOfLine - startOfLine);
+                    endOfLine - startOfLine);
                 if (lineWidth > sourceWidth) {
                     sourceWidth = lineWidth;
                 }
@@ -284,7 +325,7 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
      *
      * <p>Scrolling containers, like JScrollPane, will use this method
      * each time the user requests a unit scroll.</p>
-     * 
+     *
      * @param  visibleRect  The view area visible within the viewport
      * @param  orientation  Either SwingConstants.VERTICAL or
      *                      SwingConstants.HORIZONTAL.
@@ -294,7 +335,7 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
      * @see JScrollBar#setUnitIncrement
      */
     public int getScrollableUnitIncrement(Rectangle visibleRect,
-                                          int orientation, int direction) {
+        int orientation, int direction) {
         FontMetrics metrics = getFontMetrics(getFont());
         if (orientation == SwingConstants.VERTICAL) {
             return metrics.getHeight();
@@ -310,7 +351,7 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
      *
      * <p>Scrolling containers, like JScrollPane, will use this method
      * each time the user requests a block scroll.</p>
-     * 
+     *
      * @param  visibleRect  The view area visible within the viewport
      * @param  orientation  Either SwingConstants.VERTICAL or
      *                      SwingConstants.HORIZONTAL.
@@ -320,7 +361,7 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
      * @see JScrollBar#setBlockIncrement
      */
     public int getScrollableBlockIncrement(Rectangle visibleRect,
-                                           int orientation, int direction) {
+        int orientation, int direction) {
         Rectangle vis = getVisibleRect();
         if (orientation == SwingConstants.VERTICAL) {
             return vis.height;
@@ -338,16 +379,16 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
      * whose ancestor is a JScrollPane effectively disables horizontal
      * scrolling.
      *
-     * <p>Scrolling containers, like JViewport, will use this method each 
+     * <p>Scrolling containers, like JViewport, will use this method each
      * time they are validated.</p>
-     * 
+     *
      * @return  True if a viewport should force the Scrollable's width to
      *          match its own.
      */
     public boolean getScrollableTracksViewportWidth() {
         if (getParent() instanceof JViewport) {
-            return ((JViewport) getParent()).getWidth() >
-                getPreferredSize().width;
+            return ((JViewport) getParent()).getWidth()
+                > getPreferredSize().width;
         }
         return false;
     } // getScrollableTracksViewportWidth
@@ -358,16 +399,16 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
      * columnar text view that flowed text in left to right columns
      * could effectively disable vertical scrolling by returning true here.
      *
-     * <p>Scrolling containers, like JViewport, will use this method each 
+     * <p>Scrolling containers, like JViewport, will use this method each
      * time they are validated.</p>
-     * 
+     *
      * @return  True if a viewport should force the Scrollable's height
      *          to match its own.
      */
     public boolean getScrollableTracksViewportHeight() {
         if (getParent() instanceof JViewport) {
-            return ((JViewport) getParent()).getHeight() >
-                getPreferredSize().height;
+            return ((JViewport) getParent()).getHeight()
+                > getPreferredSize().height;
         }
         return false;
     } // getScrollableTracksViewportHeight
@@ -393,61 +434,61 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
     } // getSelectionStart
 
     /**
-     * Checks to see if the mouse if hovering over a variable; if so,
-     * attempts to find the value of that variable and set the tooltip
-     * text.
+     * Return the appropriate tooltip per the mouse position.
      *
      * @param  event  the event that caused the tooltip to appear.
      * @return  the text for the tooltip.
      */
     public String getToolTipText(MouseEvent event) {
-        // Get the current thread and stack frame.
-        Session session = SessionFrameMapper.getSessionForEvent(event);
-        ContextManager cxt = (ContextManager)
-            session.getManager(ContextManager.class);
-        ThreadReference thread = cxt.getCurrentThread();
-        if (thread != null) {
-            int frame = cxt.getCurrentFrame();
-
-            // Identify the string under the mouse.
-            Point pt = event.getPoint();
-            int offset = viewToModel(pt);
-            if (offset < 0) {
-                // out of bounds
-                return null;
-            }
-            String id = getIdentifierAt(offset);
-            if (id != null) {
-                try {
-                    // Try to interpret the value.
-                    Value value = VariableUtils.getValue(id, thread, frame);
-                    if (value == null) {
-                        return id + " = (null)";
-                    } else {
-                        return id + " = " + value.toString();
-                    }
-
-                } catch (AbsentInformationException aie) {
-                    // Catch and ignore the expected exceptions.
-                } catch (ClassNotPreparedException cnpe) {
-                } catch (FieldNotObjectException fnoe) {
-                } catch (IncompatibleThreadStateException itse) {
-                } catch (IllegalThreadStateException itse2) {
-                } catch (InvalidStackFrameException isfe) {
-                } catch (NativeMethodException nme) {
-                } catch (NoSuchFieldException nsfe) {
-                } catch (ObjectCollectedException oce) {
-                } catch (Exception e) {
-                    // Report anything out of the ordinary.
-                    System.err.println("Error reading value for " + id);
-                    e.printStackTrace();
-                }
+        Iterator iter = tooltipProducersList.iterator();
+        while (iter.hasNext()) {
+            TooltipProducer prod = (TooltipProducer) iter.next();
+            String tip = prod.produceTooltip(event, this);
+            if (tip != null) {
+                // Found a good tip, return it.
+                return tip;
             }
         }
 
+        // No one had anything to say.
         return null;
     } // getToolTipText
 
+    /**
+     * Tests if the point falls within the text area gutter.
+     *
+     * @param  x  x position of point within text area.
+     * @return  true if point is within gutter.
+     */
+    public boolean isWithinGutter(int x) {
+        return x < getGutterWidth();
+    } // isWithinGutter
+
+    /**
+     * Find the point that defines the upper-left corner of the
+     * character at the given offset.
+     *
+     * @param  offset  offset within the character buffer.
+     * @return  point in the view, or null if error.
+     */
+    public Point modelToView(int offset) {
+        int line = 0;
+        int start = 0;
+        try {
+            line = content.getLineOfOffset(offset);
+            start = content.getLineStartOffset(line);
+        } catch (BadLocationException ble) {
+            return null;
+        }
+        FontMetrics metrics = getFontMetrics(getFont());
+        Point p = new Point();
+        p.y = line * metrics.getHeight();
+        p.x = metrics.charsWidth(content.getBuffer(), start, offset - start);
+        // Can't forget to add the gutter width.
+        p.x += getGutterWidth(metrics);
+        return p;
+    } // modelToView
+
     /**
      * Invoked by Swing to draw components.
      *
@@ -458,7 +499,6 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         super.paint(g);
 
         if (!isVisible()) {
-            logCategory.report("Text area not visible in paint()");
             return;
         }
 
@@ -493,15 +533,9 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         }
 
         // Determine the width of the gutter.
-        if (gutterWidth <= 0) {
-            FontMetrics metrics = g.getFontMetrics(getFont());
-            gutterWidth = metrics.stringWidth("999999");
-        }
+        int gutterWidth = getGutterWidth();
         if (clipX + clipW > gutterWidth) {
             // The text area needs to be redrawn.
-            if (DEBUG) {
-                System.out.println("drawing text area");
-            }
             if (clipX < gutterWidth) {
                 int w = clipX + clipW - gutterWidth;
                 g.setClip(gutterWidth, clipY, w, clipH);
@@ -513,9 +547,6 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
 
         if (clipX < gutterWidth) {
             // The gutter needs to be redrawn.
-            if (DEBUG) {
-                System.out.println("drawing gutter");
-            }
             if (clipX + clipW > gutterWidth) {
                 g.setClip(clipX, clipY, gutterWidth - clipX, clipH);
             } else {
@@ -555,12 +586,7 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         for (int ii = 0; ii < gutterLayersList.size(); ii++) {
             GutterDrawLayer layer = (GutterDrawLayer) gutterLayersList.get(ii);
             if (layer.isActive()) {
-                if (DEBUG) {
-                    System.out.println("active layer: " + layer.getClass());
-                }
                 activeLayers.add(layer);
-            } else if (DEBUG) {
-                System.out.println("inactive layer: " + layer.getClass());
             }
         }
         GutterDrawLayer[] drawLayers = (GutterDrawLayer[])
@@ -570,10 +596,12 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         BasicDrawContext ctx = new BasicDrawContext(
             getForeground(), gutterColor, getFont());
 
-        if (DEBUG) {
-            System.out.println("Starting the paintGutter() loop");
-        }
+        Preferences prefs = Preferences.userRoot().node(
+            "com/bluemarsh/jswat/view");
+        boolean drawLineNums = prefs.getBoolean("viewLineNumbers",
+            Defaults.VIEW_LINE_NUMBERS);
 
+        int gutterWidth = getGutterWidth(metrics);
         int line = firstLine;
         while (line < lastLine) {
             // Have each draw layer update the draw context.
@@ -588,15 +616,16 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
             g.setColor(ctx.getBackColor());
             g.fillRect(0, y, gutterWidth, lineHeight);
 
-            // Draw in the line number for this line.
-
-            String str = Integer.toString(line + 1);
-            // Draw string so it is right-justified, but slightly
-            // away from the right border of the gutter.
-            int w = metrics.stringWidth(str);
-            y += ascent;
-            g.setColor(ctx.getForeColor());
-            g.drawString(str, gutterWidth - w - 2, y);
+            if (drawLineNums) {
+                // Draw in the line number for this line.
+                String str = Integer.toString(line + 1);
+                // Draw string so it is right-justified, but slightly
+                // away from the right border of the gutter.
+                int w = metrics.stringWidth(str);
+                y += ascent;
+                g.setColor(ctx.getForeColor());
+                g.drawString(str, gutterWidth - w - 2, y);
+            }
 
             // Advance to the next line.
             line++;
@@ -639,7 +668,6 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         try {
             nextUpdate = content.getLineStartOffset(firstLine);
         } catch (BadLocationException ble) {
-            logCategory.report("Bad location in SourceViewTextArea.paint()");
             return;
         }
 
@@ -648,12 +676,7 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         for (int ii = 0; ii < drawLayersList.size(); ii++) {
             DrawLayer layer = (DrawLayer) drawLayersList.get(ii);
             if (layer.isActive()) {
-                if (DEBUG) {
-                    System.out.println("active layer: " + layer.getClass());
-                }
                 activeLayers.add(layer);
-            } else if (DEBUG) {
-                System.out.println("inactive layer: " + layer.getClass());
             }
         }
         DrawLayer[] drawLayers = (DrawLayer[]) activeLayers.toArray(
@@ -666,10 +689,7 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         int charWidth = metrics.charWidth('a');
         char[] buffer = content.getBuffer();
 
-        if (DEBUG) {
-            System.out.println("Starting the paintText() loop");
-        }
-
+        int gutterWidth = getGutterWidth(metrics);
         int x = gutterWidth;
         int line = firstLine;
         while (line < lastLine) {
@@ -681,16 +701,10 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
             int endOfLine = 0;
             try {
                 startOfLine = content.getLineStartOffset(line);
-                endOfLine = content.getLineRealEndOffset(line);
+                endOfLine = content.getLineEndOffset(line);
             } catch (BadLocationException ble) {
-                logCategory.report(
-                    "Bad location in SourceViewTextArea.paint() while loop.");
                 return;
             }
-            if (DEBUG) {
-                System.out.println("sol = " + startOfLine + ", eol = " +
-                                   endOfLine);
-            }
 
             // Assume that we can draw until the end of the line.
             nextUpdate = endOfLine;
@@ -704,9 +718,6 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
                     nextUpdate = next;
                 }
             }
-            if (DEBUG) {
-                System.out.println("nextUpdate = " + nextUpdate);
-            }
 
             // Draw text using current context until nextUpdate.
             // Assumes that the left margin is at gutterWidth.
@@ -728,12 +739,6 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
                 // Advance to the next string draw position.
                 x += w;
             }
-            if (DEBUG) {
-                System.out.println("drawchars: @ " + lastUpdate + " + " +
-                                   updateLength + " = ''" +
-                                   new String(buffer, lastUpdate,
-                                              updateLength) + "''");
-            }
 
             if (nextUpdate == endOfLine) {
                 // See if we should draw the background color past the
@@ -762,9 +767,6 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
                 } catch (BadLocationException ble) {
                     // We're probably at the end of the view.
                 }
-                if (DEBUG) {
-                    System.out.println("advanced to line " + line);
-                }
             }
 
             // Reset the draw context to the defaults.
@@ -785,6 +787,16 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         repaint(0, vis.y, gutterWidth, vis.height);
     } // repaintGutter
 
+    /**
+     * Supports deferred automatic layout. Called when the size or internal
+     * layout of the component has changed.
+     */
+    public void revalidate() {
+        // Clear the gutter width so it is recalculated.
+        gutterWidth = 0;
+        super.revalidate();
+    } // revalidate
+
     /**
      * Selects the text found between the specified start and end locations.
      *
@@ -838,7 +850,6 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
             return content.getLength() - 1;
         }
 
-        // Fortunately our font is always mono-spaced, so this is easy.
         FontMetrics metrics = getFontMetrics(getFont());
         // Minus one because we want a zero-based line number.
         int line = viewToLine(pt) - 1;
@@ -847,13 +858,13 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
             // Use a binary search to find the character offset.
             char[] buffer = content.getBuffer();
             int lowOffset = content.getLineStartOffset(line);
-            int highOffset = content.getLineRealEndOffset(line);
+            int highOffset = content.getLineEndOffset(line);
             int midOffset = (lowOffset + highOffset) / 2;
             int textWidth = 0;
-            int offsetWidth = gutterWidth;
+            int offsetWidth = getGutterWidth(metrics);
             while (midOffset > lowOffset) {
                 textWidth = metrics.charsWidth(buffer, lowOffset,
-                                               midOffset - lowOffset);
+                    midOffset - lowOffset);
                 if (offsetWidth + textWidth < pt.x) {
                     lowOffset = midOffset;
                     offsetWidth += textWidth;
@@ -878,21 +889,25 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         /** First place the mouse was pressed. A value of -1 indicates
          * that selection is not in progress (and mouse drag events
          * should be ignored). */
-        protected int firstPos = -1;
+        private int firstPos = -1;
 
         /**
          * Invoked when the mouse has been clicked on a component.
+         *
+         * @param  e  mouse event.
          */
         public void mouseClicked(MouseEvent e) {
             // ignored
         } // mouseClicked
 
         /**
-         * Invoked when a mouse button is pressed on a component and then 
+         * Invoked when a mouse button is pressed on a component and then
          * dragged. Mouse drag events will continue to be delivered to the
          * component where the first originated until the mouse button is
          * released (regardless of whether the mouse position is within the
          * bounds of the component).
+         *
+         * @param  e  mouse event.
          */
         public void mouseDragged(MouseEvent e) {
             if (firstPos >= 0 && !e.isConsumed()) {
@@ -919,6 +934,8 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
 
         /**
          * Invoked when the mouse enters a component.
+         *
+         * @param  e  mouse event.
          */
         public void mouseEntered(MouseEvent e) {
             // ignored
@@ -926,6 +943,8 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
 
         /**
          * Invoked when the mouse exits a component.
+         *
+         * @param  e  mouse event.
          */
         public void mouseExited(MouseEvent e) {
             // ignored
@@ -934,6 +953,8 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
         /**
          * Invoked when the mouse button has been moved on a component
          * (with no buttons down).
+         *
+         * @param  e  mouse event.
          */
         public void mouseMoved(MouseEvent e) {
             // ignored
@@ -941,8 +962,12 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
 
         /**
          * Invoked when a mouse button has been pressed on a component.
+         *
+         * @param  e  mouse event.
          */
         public void mousePressed(MouseEvent e) {
+            // Put the focus in the text area.
+            SourceViewTextArea.this.requestFocus();
             if (!e.isConsumed() && !e.isPopupTrigger()) {
                 firstPos = viewToModel(e.getPoint());
                 if (firstPos < 0) {
@@ -960,6 +985,8 @@ public class SourceViewTextArea extends JComponent implements Scrollable, Source
 
         /**
          * Invoked when a mouse button has been released on a component.
+         *
+         * @param  e  mouse event.
          */
         public void mouseReleased(MouseEvent e) {
             // ignored
